from pathlib import Path
from math import sin, pi, cos

from newsflash.widgets import LineChart
from ..snapshot_test import SnapshotTest


class SineLine(SnapshotTest):
    name: str = "Sine Line"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "sine-line.html"
    description: str = """
        Expected to see a line chart plotting a sine wave from 0 to 2 pi. The width 
        of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = LineChart(
            width=600,
            height=400,
        )

        chart.title = "Sine Line Chart"

        xs = [x / 100 * 2 * pi for x in range(101)]
        ys = [sin(x) for x in xs]

        chart.set_values(xs=xs, ys=ys)

        return chart._render_update()


class CosineLine(SnapshotTest):
    name: str = "Cosine Line"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "cosine-line.html"
    description: str = """
        Expected to see a line chart plotting a cosine wave for 10 periods. The amplitude
        of the wave is 10. The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = LineChart(
            width=600,
            height=400,
        )

        chart.title = "Cosine Line Chart"

        xs = [x / 500 * 20 * pi for x in range(501)]
        ys = [10 * cos(x) for x in xs]

        chart.set_values(xs=xs, ys=ys)

        return chart._render_update()


class SineLineWithCustomLabels(SnapshotTest):
    name: str = "Sine Line Custom Labels"
    path_to_rendered: Path = (
        Path(__file__).parent / "rendered" / "sine-line-with-custom-labels.html"
    )
    description: str = """
        Expected to see a line chart plotting a sine wave for 10 periods. The amplitude
        of the wave is 10. The chart has x labels at every interval of 2*pi. 
        The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = LineChart(
            width=600,
            height=400,
        )

        chart.title = "Sine Line With Custom Labels"

        xs = [x / 500 * 20 * pi for x in range(501)]
        ys = [10 * sin(x) for x in xs]
        x_labels: dict[float, str] = {
            0: "0",
            2 * pi: "2π",
            4 * pi: "4π",
            6 * pi: "6π",
            8 * pi: "8π",
            10 * pi: "10π",
            12 * pi: "12π",
            14 * pi: "14π",
            16 * pi: "16π",
            18 * pi: "18π",
            20 * pi: "20π",
        }

        chart.set_values(xs=xs, ys=ys, x_labels=x_labels)

        return chart._render_update()


class SineLineWithCustomXRange(SnapshotTest):
    name: str = "Sine Line Custom X Range"
    path_to_rendered: Path = (
        Path(__file__).parent / "rendered" / "sine-line-custom-x-range.html"
    )
    description: str = """
        Expected to see a line chart plotting a sine wave for 1 period. The amplitude
        of the wave is 1. The x-axis range is from -2π to 4π. x-labels are set at 0 and 2π.
        The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = LineChart(
            width=600,
            height=400,
        )

        chart.title = "Sine Line With Custom X Range"

        xs = [x / 100 * 2 * pi for x in range(101)]
        ys = [sin(x) for x in xs]
        x_labels: dict[float, str] = {
            0: "0",
            2 * pi: "2π",
        }

        chart.set_values(
            xs=xs, ys=ys, min_x_value=-2 * pi, max_x_value=4 * pi, x_labels=x_labels
        )

        return chart._render_update()



# # from newsflash.endpoints.callback import _get_callback_inputs
# from newsflash.widgets.widgets import Widget
# from newsflash.endpoints.parsers import RequestValues


# class DummyInputWidget(Widget):
#     pass


# class DummyWidget(Widget):
#     _callback_fn_name: str = "test_callback"

#     def test_callback(self, input_widget: DummyInputWidget) -> list[Widget]: ...


# def test_get_callback_inputs():
#     request_values = RequestValues(
#         url_path="/test",
#         trigger_element_id="test-trigger",
#         widget_attributes={},
#     )

#     widget = DummyWidget(
#         request_values=request_values,
#     )
#     widget._post_init()

#     result = widget._get_callback_inputs()

#     assert "input_widget" in result



from unittest.mock import patch

from newsflash.widgets.widgets import WidgetContainer
from newsflash.widgets.charts import Chart


def test_default_on_load():
    chart = Chart()
    assert chart.on_load() == []


def test_render_container():
    class DummyChart(Chart):
        id: str = "dummy-chart"

    chart = DummyChart()

    def new_render(self, *args, **kwargs) -> str:
        return f"<div id={self.widget_id}></div>"

    with patch.object(WidgetContainer, "render", new=new_render):
        rendered = chart.render()

    assert rendered == "<div id=dummy-chart></div>"



from fontTools.ttLib import TTFont

from newsflash.svg.box import Box
from newsflash.svg.utils import Point
from newsflash.svg.utils.fonts import lora
from newsflash.svg.element import ElementGroup
from newsflash.svg.elements import build_rectangle_from_bottom_center

from .xy_chart import build_xy_chart
from .axes import (
    AxesConfig,
    build_y_axis_config,
    build_x_axis_config_barchart,
)


def build_bars(
    bars: list[float] | list[int],
    chart_box: Box,
    xs: list[float] | list[int] | None = None,
) -> ElementGroup:
    assert xs is None or len(xs) == len(bars)
    elements = ElementGroup()

    # TODO: determine x_width inside this if statement
    # is a little bit hacky
    if xs is None:
        xs = list(range(len(bars)))
        x_width = (max(xs) - min(xs) + 1) / len(xs)
    else:
        x_width = (max(xs) - min(xs)) / len(xs)

    for x, y in zip(xs, bars):
        rect = build_rectangle_from_bottom_center(
            bottom_center=Point(x=x, y=0.0),
            width=0.9 * x_width,
            height=y,
            rounded=0.05,
            classes=["bar"],
            box=chart_box,
        )
        elements.append(rect)

    return elements


def build_barchart(
    values: list[float] | list[int],
    labels: list[str],
    width: float,
    height: float,
    title: str,
    font: TTFont = lora,
    title_font_size: int = 32,
    label_font_size: int = 16,
) -> ElementGroup:
    barchart_elements = ElementGroup()

    x_padding = 0.5
    axes = AxesConfig(
        x=build_x_axis_config_barchart(labels=labels),
        y=build_y_axis_config(values=values, min_value=0),
    )

    barchart_elements, chart_box = build_xy_chart(
        axes=axes,
        width=width,
        height=height,
        title=title,
        x_padding=x_padding,
        font=font,
        title_font_size=title_font_size,
        label_font_size=label_font_size,
    )

    bars = build_bars(bars=values, chart_box=chart_box)
    barchart_elements.extend(bars)

    return barchart_elements



from typing import Type, TypeVar
from pathlib import Path

from fastapi import FastAPI, Request
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse

from newsflash.widgets.widgets import Widget
from newsflash.widgets import Notifications
from newsflash.templates.templates import template_registry
from newsflash.endpoints.parsers import RequestValues, parse_request_values
from newsflash.endpoints.callback import get_page_callback


W = TypeVar("W", bound=Widget)


class Page(Widget):
    path: str
    title: str
    query_params: dict[str, list[str]] = {}

    def _post_init(self) -> None:
        self.children.append(Notifications())  # Always add Notification widget to pages

        return super()._post_init()


class App(FastAPI):
    pages: dict[str, Page]

    def __init__(
        self,
        pages: list[Page],
        template_folders: list[tuple[str, Path]],
        static_folders: list[tuple[str, Path]] = [],
    ) -> None:
        super().__init__()

        self.pages = {page.path: page for page in pages}

        for folder in template_folders:
            template_registry.register_template_folder(folder[0], folder[1])

        static_folders.append(
            ("/_newsflash/static", Path(__file__).parent / "assets" / "staticfiles")
        )
        self._mount_static_folders(static_folders)
        self._build_page_endpoints()
        self._build_callback_endpoints()

    def get_widget(
        self,
        path: str,
        type: Type[W],
        id: str | None = None,
        request_values: RequestValues | None = None,
    ) -> W:
        page = self.pages[path].model_copy(deep=True)
        # assert id.startswith(f"{page.id}/"), (
        #     f"Widget id '{id}' not found on page '{page.id}'"
        # )
        if id is not None:
            id = id.removeprefix(f"{page.id}/")

        page._post_init()
        return page.get_child_widget(type=type, id=id, request_values=request_values)

    def _mount_static_folders(self, static_folders: list[tuple[str, Path]]) -> None:
        for mount_path, directory in static_folders:
            self.mount(
                mount_path,
                StaticFiles(directory=directory),
            )

    def _build_page_endpoints(self):
        for page_path, page in self.pages.items():
            assert page.template is not None, "Page template is not set."

            page_endpoint = get_page_callback(page)
            self.add_api_route(
                page_path, page_endpoint, methods=["GET"], response_class=HTMLResponse
            )

    def _build_callback_endpoints(self):
        async def callback_endpoint(request: Request, widget_id: str) -> str:
            body = await request.form()
            headers = request.headers
            request_values = parse_request_values(body, headers)

            widget = self.get_widget(
                path=request_values.url_path,
                type=Widget,
                id=widget_id,
                request_values=request_values,
            )

            widgets_to_render = widget._call_callback()

            rendered_widgets = []
            assert isinstance(widgets_to_render, list), (
                "Callback must return a list of widgets"
            )
            for widget in widgets_to_render:
                assert isinstance(widget, Widget), (
                    "Callback must return a list of widgets"
                )
                widget.hx_swap_oob = True
                rendered_widgets.append(widget._render_update())

            return "\n".join(rendered_widgets)

        self.add_api_route(
            "/{widget_id:path}",
            callback_endpoint,
            methods=["POST"],
            response_class=HTMLResponse,
        )



from typing import Any

from newsflash.svg.element import ElementGroup
from newsflash.svg.charts.barchart import build_barchart
from newsflash.svg.charts.linechart import build_linechart
from newsflash.svg.charts.histogram import build_histogram

from .widgets import Widget, WidgetContainer


class Chart(Widget):
    template: tuple[str, str] = ("svg", "svg.svg")
    width: float = 0.0
    height: float = 0.0
    title: str = ""
    title_font_size: int = 32
    label_font_size: int = 16

    elements: ElementGroup = ElementGroup()
    styles: dict[str, str] = {
        "position": "absolute",
    }

    include_in_context: set[str] = {
        "id",
        "width",
        "height",
        "hx_swap_oob",
        "classes",
        "styles",
        "attributes",
    }

    _values_from_request: list[str] = ["width", "height"]
    _callback_fn_name: str = "on_load"

    def on_load(self, *args, **kwargs) -> list[Widget]:
        """Event handler for chart load events."""
        return []

    def render(self) -> str:
        container = WidgetContainer(
            widget_id=self.id,
        )
        container.hx_include.extend(self.hx_include)

        context = self.get_additional_context()
        return container.render(additional_context=context)

    def _render_update(self) -> str:
        return super().render()

    def get_additional_context(self) -> dict[str, Any]:
        additional_context = super().get_additional_context()
        additional_context.update(
            {
                "content": "\n    ".join(
                    element.render() for element in self.elements.elements
                ),
            }
        )
        return additional_context


class BarChart(Chart):
    def set_values(
        self,
        values: list[float] | list[int],
        labels: list[str],
    ) -> None:
        self.elements = build_barchart(
            values=values,
            labels=labels,
            width=self.width,
            height=self.height,
            title=self.title,
            title_font_size=self.title_font_size,
            label_font_size=self.label_font_size,
        )


class LineChart(Chart):
    classes: list[str] = ["line-chart"]

    def set_values(
        self,
        xs: list[float] | list[int],
        ys: list[float] | list[int],
        # TODO: fix type annotation
        x_labels: dict[Any, Any] | None = None,
        min_x_value: float | int | None = None,
        max_x_value: float | int | None = None,
    ) -> None:
        self.elements = build_linechart(
            xs=xs,
            ys=ys,
            width=self.width,
            height=self.height,
            title=self.title,
            x_labels=x_labels,
            min_x_value=min_x_value,
            max_x_value=max_x_value,
            title_font_size=self.title_font_size,
            label_font_size=self.label_font_size,
        )


class Histogram(Chart):
    def set_values(
        self,
        values: list[float] | list[int],
        num_bins: int,
    ) -> None:
        self.elements = build_histogram(
            values=values,
            num_bins=num_bins,
            width=self.width,
            height=self.height,
            title=self.title,
            title_font_size=self.title_font_size,
            label_font_size=self.label_font_size,
        )



from pathlib import Path

from newsflash.widgets import BarChart
from ..snapshot_test import SnapshotTest


class AscendingBarChart5(SnapshotTest):
    name: str = "Ascending Bar Chart 5"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "ascending-5.html"
    description: str = """
        Expected to see five bars labelled A to E with heights increasing from 
        left to right in steps of 10. The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = BarChart(
            width=600,
            height=400,
        )

        chart.title = "Ascending Bar Chart"

        chart.set_values(
            values=[10, 20, 30, 40, 50],
            labels=["A", "B", "C", "D", "E"],
        )

        return chart._render_update()


class AscendingBarChart20(SnapshotTest):
    name: str = "Ascending Bar Chart 20"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "ascending-20.html"
    description: str = """
        Expected to see twenty bars labelled 1 to 20 with heights increasing from 
        left to right in steps of 5. The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = BarChart(
            width=600,
            height=400,
        )

        chart.title = "Ascending Bar Chart"

        chart.set_values(
            values=[x * 5 for x in range(1, 21)],
            labels=[str(x) for x in range(1, 21)],
        )

        return chart._render_update()



from newsflash.svg.box import Box
from newsflash.svg.utils import Point
from newsflash.svg.utils.fonts import font_size_to_height, lora
from newsflash.svg.elements import build_text, Text, build_line
from newsflash.svg.element import ElementGroup


def build_title_box(width: float, title_font_size: int, label_font_size: int) -> Box:
    return Box(top=0, right=width, bottom=font_size_to_height(title_font_size), left=0)


def build_title_text(box: Box, title: str, font_size: int) -> Text:
    return build_text(
        pos=Point(x=0.5, y=0),
        text=title,
        font=lora,
        font_size=font_size,
        horizontal_align="center",
        vertical_align="top",
        box=box,
        classes=["text"],
    )


def build_multipler_text(multiplier: float, font_size: int, box: Box) -> Text:
    return build_text(
        pos=Point(x=0.0, y=1.0),
        text=f"× {multiplier:,}",
        font=lora,
        font_size=font_size,
        horizontal_align="left",
        vertical_align="bottom",
        box=box,
        classes=["text"],
    )


def _compute_x_axis_height(font_size: int) -> float:
    return font_size_to_height(font_size) * 1.75


def build_y_axis_box(
    min_value: float,
    max_value: float,
    title_box: Box,
    max_label_width: float,
    svg_height: float,
    label_font_size: int,
) -> Box:
    x_axis_height = _compute_x_axis_height(label_font_size)

    return Box(
        top=title_box.bottom,
        right=max_label_width + label_font_size,
        bottom=svg_height - x_axis_height,
        left=0,
        top_y_value=max_value,
        bottom_y_value=min_value,
    )


def build_y_axis(
    y_labels: list[str],
    y_label_positions: list[float] | list[int],
    font_size: int,
    y_axis_box: Box,
) -> ElementGroup:
    elements = ElementGroup()

    for y_label, y_label_position in zip(y_labels, y_label_positions):
        label = build_text(
            pos=Point(x=0, y=y_label_position),
            text=str(y_label),
            font=lora,
            font_size=font_size,
            horizontal_align="left",
            vertical_align="middle",
            box=y_axis_box,
            classes=["text"],
        )
        elements.append(label)

    return elements


def build_x_axis_box(
    svg_height: float,
    svg_width: float,
    font_size: int,
    y_axis_box: Box,
    min_x_axis_value: float,
    max_x_axis_value: float,
    padding_left: float = 0.0,
    padding_right: float = 0.0,
) -> Box:
    return Box(
        top=svg_height - _compute_x_axis_height(font_size),
        right=svg_width,
        bottom=svg_height,
        left=y_axis_box.right,
        top_y_value=1.0,
        bottom_y_value=0.0,
        right_x_value=max_x_axis_value,
        left_x_value=min_x_axis_value,
        padding_left=padding_left,
        padding_right=padding_right,
    )


def build_x_axis(
    labels: list[str],
    label_positions: list[float] | list[int],
    font_size: int,
    x_axis_box: Box,
) -> ElementGroup:
    elements = ElementGroup()

    assert len(labels) == len(label_positions)

    for label_position, label_value in zip(label_positions, labels):
        label = build_text(
            pos=Point(x=label_position, y=0),
            text=label_value,
            font=lora,
            font_size=font_size,
            horizontal_align="center",
            vertical_align="bottom",
            box=x_axis_box,
            classes=["text"],
        )
        elements.append(label)

    return elements


def build_chart_box(
    svg_width: float,
    min_x_axis_value: float,
    max_x_axis_value: float,
    min_y_axis_value: float,
    max_y_axis_value: float,
    title_box: Box,
    x_axis_box: Box,
    y_axis_box: Box,
    padding_left: float = 0.0,
    padding_right: float = 0.0,
) -> Box:
    return Box(
        top=title_box.bottom,
        right=svg_width,
        bottom=x_axis_box.top,
        left=y_axis_box.right,
        top_y_value=max_y_axis_value,
        right_x_value=max_x_axis_value,
        bottom_y_value=min_y_axis_value,
        left_x_value=min_x_axis_value,
        padding_left=padding_left,
        padding_right=padding_right,
    )


def build_horizontal_grid_lines(
    y_labels: list[float] | list[int],
    chart_box: Box,
    padding_left: float = 0.0,
    padding_right: float = 0.0,
) -> ElementGroup:
    elements = ElementGroup()

    for y_label in y_labels:
        line = build_line(
            from_pos=Point(x=chart_box.left_x_value - padding_left, y=y_label),
            to_pos=Point(x=chart_box.right_x_value + padding_right, y=y_label),
            box=chart_box,
            classes=["stroke-white", "stroke-rounded", "stroke-width-2", "grid-line"],
        )
        elements.append(line)

    return elements



from newsflash.widgets.text import Notifications


def test_notifications_push():
    notifications_widget = Notifications()

    # Push a notification
    notifications_widget.push("Test notification", level="info", duration=3000)

    # Check if the notification was added
    assert len(notifications_widget.notifications) == 1
    assert notifications_widget.notifications[0].message == "Test notification"
    assert notifications_widget.notifications[0].level == "info"
    assert notifications_widget.notifications[0].duration == 3000

    # Push another notification
    notifications_widget.push("Another notification", level="warning")

    # Check if the second notification was added
    assert len(notifications_widget.notifications) == 2
    assert notifications_widget.notifications[1].message == "Another notification"
    assert notifications_widget.notifications[1].level == "warning"
    assert notifications_widget.notifications[1].duration == 5000  # Default duration



from typing import Any

from pytest import raises

from newsflash.endpoints.parsers import RequestValues
from newsflash.widgets.widgets import Widget


class DummyWidget(Widget):
    id: str = "dummy-widget"
    test_value_a: str = ""
    test_value_b: float = 0.0

    _values_from_request: list[str] = ["test_value_a", "test_value_b"]
    _callback_fn_name: str = "dummy_callback"

    def dummy_callback(self) -> list[Widget]: ...


def test_widget_set_value_from_request():
    dummy_widget = DummyWidget()

    request_data: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_b": 3.14,
    }

    dummy_widget._set_value_from_request("test_value_a", request_data)
    dummy_widget._set_value_from_request("test_value_b", request_data)

    assert dummy_widget.test_value_a == "Hello, World!"
    assert dummy_widget.test_value_b == 3.14


def test_widget_set_value_no_attribute():
    dummy_widget = DummyWidget()

    request_data: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_c": 3.14,
    }

    with raises(AssertionError) as e:
        dummy_widget._set_value_from_request("test_value_c", request_data)

    assert str(e.value) == "Widget has no attribute 'test_value_c'"


# TODO: design expected behavior for missing keys
# def test_widget_set_value_from_request_missing_key():
#     dummy_widget = DummyWidget()
#     request_data: dict[str, Any] = {
#         "dummy-widget-test_value_a": "Hello, World!",
#         # "dummy-widget-test_value_b" is missing
#     }

#     with raises(AssertionError) as e:
#         dummy_widget._set_value_from_request("test_value_b", request_data)

#     assert str(e.value) == "No value provided for key 'test_value_b'"


def test_widget_set_value_wrong_type():
    dummy_widget = DummyWidget()

    request_data: dict[str, Any] = {
        "dummy-widget-test_value_a": 100,
        "dummy-widget-test_value_b": 3.14,
    }

    with raises(AssertionError) as e:
        dummy_widget._set_value_from_request("test_value_a", request_data)

    assert (
        str(e.value)
        == "Expected type <class 'str'> for key 'test_value_a', got <class 'int'>"
    )


def test_widget_set_values_from_request():
    dummy_widget = DummyWidget()

    request_data: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_b": 3.14,
    }

    request_values = RequestValues(
        trigger_element_id="some-widget",
        url_path="/dummy",
        widget_attributes=request_data,
    )

    dummy_widget._set_values_from_request(request_values)

    assert dummy_widget.test_value_a == "Hello, World!"
    assert dummy_widget.test_value_b == 3.14


def test_widget_factory():
    request_data: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_b": 3.14,
    }

    request_values = RequestValues(
        trigger_element_id="some-widget",
        url_path="/dummy",
        widget_attributes=request_data,
    )

    widget = DummyWidget(request_values=request_values)
    widget._post_init()

    assert isinstance(widget, DummyWidget)
    assert widget.test_value_a == "Hello, World!"
    assert widget.test_value_b == 3.14


def test_get_widget_callback_fn():
    dummy_widget = DummyWidget()
    callback_fn = dummy_widget._get_callback_fn()
    assert callback_fn == dummy_widget.dummy_callback


def test_get_widget_callback_none():
    class NoCallbackWidget(Widget):
        id: str = "no-callback-widget"
        _callback_fn_name: str | None = None

    no_callback_widget = NoCallbackWidget()

    callback_fn = no_callback_widget._get_callback_fn()
    assert callback_fn is None


def test_get_widget_callback_fn_no_callback():
    class NoCallbackWidget(Widget):
        id: str = "no-callback-widget"
        _callback_fn_name: str = "non_existent_callback"

    no_callback_widget = NoCallbackWidget()

    with raises(AssertionError) as e:
        no_callback_widget._get_callback_fn()

    assert str(e.value) == "Widget has no callback function 'non_existent_callback'"



from .widgets import Widget


class HTML(Widget):
    template: tuple[str, str] = ("widgets", "html.html")
    html_content: str = ""

    _values_from_request: set[str] = {
        "html_content",
    }

    include_in_context: set[str] = {
        "id",
        "hx_swap_oob",
        "html_content",
    }



from newsflash.widgets.widgets import Widget


class DummyWidgetA(Widget):
    id: str = "widget-a"


class DummyWidgetB(Widget):
    id: str = "widget-b"


class DummyWidgetC(Widget):
    id: str = "widget-c"

    _callback_fn_name: str = "dummy_callback"

    def dummy_callback(
        self,
        widget_a: DummyWidgetA,
        widget_b: DummyWidgetB,
    ) -> list[Widget]: ...

    def render(self) -> str: ...


def test_build_hx_include():
    widget = DummyWidgetC()
    widget._build_hx_include()

    assert isinstance(widget.hx_include, list)
    assert len(widget.hx_include) == 2
    assert "#widget-a" in widget.hx_include
    assert "#widget-b" in widget.hx_include


class DummyWidgetD(Widget):
    id: str = "widget-d"

    _callback_fn_name: str = "dummy_callback_no_inputs"

    def dummy_callback_no_inputs(self) -> list[Widget]: ...


def test_build_hx_include_no_inputs():
    widget = DummyWidgetD()
    widget._build_hx_include()

    assert isinstance(widget.hx_include, list)
    assert len(widget.hx_include) == 0


# TODO: decide on behavior for illegal inputs

# class DummyWidgetE(Widget):
#     id: str = "widget-e"

#     _callback_fn_name: str = "dummy_callback_illegal_inputs"

#     def dummy_callback_illegal_inputs(
#         self,
#         not_a_widget: float,
#     ) -> list[Widget]: ...


# def test_build_hx_include_illegal_inputs():
#     widget = DummyWidgetE()

#     with raises(AssertionError):
#         widget._build_hx_include()


class DummyWidgetF(Widget):
    id: str = "widget-f"

    _callback_fn_name: str = "dummy_callback"

    def dummy_callback(
        self,
        widget_a: DummyWidgetA,
        widget_b: DummyWidgetB,
        widget_c: DummyWidgetC,
    ) -> list[Widget]: ...

    def render(self) -> str:
        return f"Dummy Widget F, with hx_include: {', '.join(self.hx_include)}"


def test_build_rendered_widgets():
    widget = DummyWidgetF()
    widget._build_hx_include()

    rendered_widget = widget._render_update()
    expected = "Dummy Widget F, with hx_include: #widget-a, #widget-b, #widget-c"
    assert rendered_widget == expected



from newsflash.endpoints.parsers import (
    _parse_trigger_element_id,
    _parse_url_path,
    _parse_chart_dimensions,
    parse_request_values,
)
from starlette.datastructures import Headers, FormData


def test_parse_trigger_element_id():
    headers = Headers(
        {
            "hx-trigger": "my-button-id",
        }
    )

    trigger_element_id = _parse_trigger_element_id(headers)
    assert trigger_element_id == "my-button-id"


def test_parse_url_path():
    headers = Headers(
        {
            "hx-current-url": "https://example.com/some/path?query=param",
        }
    )

    url_path = _parse_url_path(headers)
    assert url_path == "/some/path"


def test_parse_chart_dimensions():
    body = FormData(
        {
            "dimensions": '{"chart-width": "800", "chart-height": "600"}',
        }
    )

    dimensions = _parse_chart_dimensions(body)
    assert dimensions == {"chart-width": 800.0, "chart-height": 600.0}


def test_parse_multiple_chart_dimensions():
    body = FormData(
        {
            "dimensions": '{"chart-1-width": "800", "chart-1-height": "600", '
            '"chart-2-width": "400", "chart-2-height": "300"}',
        }
    )

    dimensions = _parse_chart_dimensions(body)
    assert dimensions == {
        "chart-1-width": 800.0,
        "chart-1-height": 600.0,
        "chart-2-width": 400.0,
        "chart-2-height": 300.0,
    }


def test_parse_chart_dimensions_no_dimensions():
    body = FormData({})

    dimensions = _parse_chart_dimensions(body)
    assert dimensions == {}


def test_parse_request_values():
    body = FormData(
        {
            "input-1-value": "some text",
            "select-1-value": "option-2",
            "dimensions": '{"chart-width": "800", "chart-height": "600"}',
        }
    )
    headers = Headers(
        {
            "hx-trigger": "input-1",
            "hx-current-url": "https://example.com/dashboard/view?query=value",
        }
    )

    request_values = parse_request_values(body, headers)

    assert request_values.trigger_element_id == "input-1"
    assert request_values.url_path == "/dashboard/view"
    assert request_values.widget_attributes == {
        "input-1-value": "some text",
        "select-1-value": "option-2",
        "chart-width": 800.0,
        "chart-height": 600.0,
    }



from typing import Any

from pydantic import BaseModel, ConfigDict

from newsflash.templates.templates import template_registry


class Element(BaseModel):
    template: tuple[str, str] | None = None
    id: str = ""

    include_in_context: set[str] = set()

    classes: list[str] = []
    styles: list[str] = []
    attributes: dict[str, str] = {}

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def get_additional_context(self) -> dict[str, Any]:
        return {}

    def render(self, additional_context: dict[str, Any] = {}) -> str:
        context = self.model_dump(include=self.include_in_context)
        context.update(self.get_additional_context())
        context.update(additional_context)

        assert self.template is not None, (
            f"Element with ID '{self.id}' has no template defined"
        )

        template = template_registry.get_template(
            template_folder=self.template[0], template_name=self.template[1]
        )

        rendered = template.render(context)
        return rendered


class ElementGroup(Element):
    elements: list[Element] = []

    def append(self, element: Element) -> None:
        self.elements.append(element)

    def extend(self, other: "ElementGroup") -> None:
        self.elements.extend(other.elements)

    def get_additional_context(self) -> dict[str, Any]:
        return {
            "content": "\n    ".join(element.render() for element in self.elements),
        }



from pathlib import Path

from pytest import fixture, raises

from newsflash import App, Page
from newsflash.widgets.widgets import Widget


class DummyWidgetA(Widget):
    id: str = "widget_a"


class DummyWidgetB(Widget):
    id: str = "widget_b"


class DummyWidgetC(Widget):
    id: str = "widget_c"


@fixture
def test_page() -> Page:
    return Page(
        id="test-page",
        path="/test-page",
        title="Test Page",
        template=("templates", "test_page.html"),
        children=[DummyWidgetA(), DummyWidgetB()],
    )


@fixture
def test_app(test_page: Page) -> App:
    app = App(
        pages=[test_page],
        template_folders=[("templates", Path.cwd() / "dummy" / "path")],
    )
    return app


def test_get_widget_by_type(
    test_app: App,
):
    widget_a = test_app.get_widget(
        path="/test-page",
        type=DummyWidgetA,
    )
    assert isinstance(widget_a, DummyWidgetA)
    assert widget_a.id == "widget_a"

    widget_b = test_app.get_widget(path="/test-page", type=DummyWidgetB)
    assert isinstance(widget_b, DummyWidgetB)
    assert widget_b.id == "widget_b"


def test_query_one_with_id(
    test_app: App,
):
    widget_a = test_app.get_widget(
        path="/test-page",
        type=Widget,
        id="widget_a",
    )
    assert isinstance(widget_a, DummyWidgetA)
    assert widget_a.id == "widget_a"


def test_query_one_multiple_found(
    test_app: App,
):
    test_app.pages["/test-page"].children.append(DummyWidgetA(id="widget_a_2"))

    with raises(ValueError) as e:
        test_app.get_widget(
            path="/test-page",
            type=DummyWidgetA,
        )

    assert "specify an id" in str(e.value).lower()


def test_query_one_not_found(
    test_app: App,
):
    with raises(ValueError) as e:
        test_app.get_widget(
            path="/test-page",
            type=DummyWidgetC,
        )

    assert "no widgets" in str(e.value).lower()



from pydantic import BaseModel

from .widgets import Widget


class ValueDisplay(Widget):
    template: tuple[str, str] = ("widgets", "value_display.html")
    label: str = ""
    value: str = ""

    include_in_context: set[str] = {
        "id",
        "value",
        "hx_include",
        "hx_swap_oob",
        "label",
        "value",
    }

    _values_from_request: list[str] = ["label", "value"]


class Paragraph(Widget):
    template: tuple[str, str] = ("widgets", "paragraph.html")
    text: str = ""

    include_in_context: set[str] = {"id", "value", "hx_include", "hx_swap_oob", "text"}

    _values_from_request: list[str] = ["text"]


class Notification(BaseModel):
    message: str = ""
    level: str
    duration: int  # duration in milliseconds


class Notifications(Widget):
    id: str = "newsflash-notifications-container"
    template: tuple[str, str] = ("widgets", "notifications.html")
    notifications: list[Notification] = []  # List of notifications to display

    include_in_context: set[str] = {
        "notifications",
    }

    def push(self, message: str, level: str = "info", duration: int = 5000) -> None:
        notification = Notification(
            message=message,
            level=level,
            duration=duration,
        )
        self.notifications.append(notification)



from typing import Callable

from .widgets import Widget


class Input(Widget):
    template: tuple[str, str] = ("widgets", "input.html")
    value: str | None = None
    default: Callable[[], str] | None = None
    autofocus: bool = False
    type: str = "text"
    placeholder: str = ""

    include_in_context: set[str] = {
        "id",
        "value",
        "hx_include",
        "hx_swap_oob",
        "full_path",
        "autofocus",
        "type",
        "placeholder",
    }

    _values_from_request: list[str] = ["value"]
    _callback_fn_name: str = "on_input"

    def _post_init(self) -> None:
        if self.value is None:
            if self.default is not None:
                self.value = self.default()
            else:
                self.value = ""

        super()._post_init()

    def on_input(self, *args, **kwargs) -> list[Widget]:
        """Event handler for input events."""
        return []


class TextArea(Input):
    template: tuple[str, str] = ("widgets", "textarea.html")
    rows: int = 7
    spellcheck: bool = False

    include_in_context: set[str] = {
        "id",
        "hx_include",
        "hx_swap_oob",
        "full_path",
        "autofocus",
        "value",
        "rows",
        "spellcheck",
        "placeholder",
    }


class Select(Widget):
    template: tuple[str, str] = ("widgets", "select.html")
    options: list[str] = ["-"]
    selected: str | None = None
    default: Callable[[], str] | None = None

    include_in_context: set[str] = {
        "id",
        "hx_include",
        "hx_swap_oob",
        "full_path",
        "options",
        "selected",
    }

    _callback_fn_name: str = "on_select"
    _values_from_request: list[str] = ["selected"]

    def _post_init(self) -> None:
        if self.selected is None:
            if self.default is not None:
                self.selected = self.default()
            elif len(self.options) > 0:
                self.selected = self.options[0]
            else:
                raise ValueError("Select widget has no options to select from.")

        super()._post_init()

    def on_select(self, *args, **kwargs) -> list[Widget]:
        """Event handler for select events."""
        return []


class Button(Widget):
    template: tuple[str, str] = ("widgets", "button.html")
    label: str = "Click Me"
    hx_include: list[str] = []
    disabled: bool = False
    classes: list[str] = ["newsflash-button"]

    include_in_context: set[str] = {
        "id",
        "hx_include",
        "hx_swap_oob",
        "full_path",
        "label",
        "disabled",
        "classes",
    }

    _callback_fn_name: str = "on_click"

    def on_click(self, *args, **kwargs) -> list[Widget]:
        """Event handler for button click events."""
        return []



from pathlib import Path

from pytest import fixture
from fontTools.ttLib import TTFont

from newsflash.svg.utils.fonts import (
    get_text_width,
    get_text_height,
    font_size_to_height,
)


@fixture
def font() -> TTFont:
    path_to_font = (
        Path(__file__).parent.parent.parent.parent.parent
        / "newsflash"
        / "assets"
        / "fonts"
        / "lora"
        / "Lora-Regular.ttf"
    )
    return TTFont(path_to_font)


def test_get_text_width(font: TTFont):
    text = "Hello, World!"
    font_size = 12

    width = get_text_width(font, text, font_size)
    assert isinstance(width, float)
    assert width > 0


def test_get_text_height(font: TTFont):
    text = "Hello, World!"
    font_size = 12

    y_min, y_max = get_text_height(font, text, font_size)
    assert isinstance(y_min, float)
    assert isinstance(y_max, float)
    assert y_min < y_max


def test_font_size_to_height():
    assert font_size_to_height(15) == 20
    assert font_size_to_height(24) == 32



import pytest

from newsflash.svg.charts.utils import (
    order_of_magnitude,
    n_significant_figures,
    order_of_magnitude_for_significant_figures,
    smart_round,
    get_step_decimal_places,
)


@pytest.mark.parametrize(
    "value, expected",
    [
        (0, 0),
        (5, 0),
        (75, 1),
        (500.2, 2),
        (0.5111, -1),
        (0.023, -2),
        (-7800, 3),
    ],
)
def test_order_of_magnitude(value: int, expected: int) -> None:
    assert order_of_magnitude(value) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (0.25, 2),
        (2, 1),
        (200, 1),
        (0.00456, 3),
        (12345.6789, 9),
    ],
)
def test_n_significant_figures(value: float, expected: int) -> None:
    assert n_significant_figures(value) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (0.25, -2),
        (2, 0),
        (200, 2),
        (0.00456, -5),
        (12345.6789, -4),
    ],
)
def test_order_of_magnitude_for_significant_figures(
    value: float, expected: int
) -> None:
    assert order_of_magnitude_for_significant_figures(value) == expected


@pytest.mark.parametrize(
    "numbers, expected_rounded, expected_decimal_places",
    [
        (
            [0.0, 0.25, 0.5, 0.75, 1.0],
            [0.0, 0.2, 0.5, 0.8, 1.0],
            1,
        ),
        (
            [0, 0.1, 1.01, 2.2, 3, 4],
            [0, 0.1, 1, 2.2, 3, 4],
            1,
        ),
    ],
)
def test_smart_round(
    numbers: list[float],
    expected_rounded: list[float],
    expected_decimal_places: int,
) -> None:
    rounded, decimal_places = smart_round(numbers)
    assert rounded == expected_rounded
    assert decimal_places == expected_decimal_places


@pytest.mark.parametrize(
    "numbers, expected_decimal_places",
    [
        (
            [0.0, 0.25, 0.5, 0.75, 1.0],
            2,  # because 0.25 requires 2 decimal places
        ),
        (
            [0, 0.1, 1.01, 2.2, 3, 4],
            1,  # because 0.1 requires 1 decimal place
        ),
    ],
)
def test_get_step_decimal_places(
    numbers: list[float],
    expected_decimal_places: int,
) -> None:
    decimal_places = get_step_decimal_places(numbers)
    assert decimal_places == expected_decimal_places



from .all_tests import get_all_tests


def test_snapshots():
    for snapshot_test in get_all_tests().values():
        assert snapshot_test.passes_test(), (
            "Snapshot test failed. Please run the snapshot test approval app and review the changes."
        )



from pydantic import BaseModel

from .utils.point import XCoor, YCoor, Point


class Box(BaseModel):
    top: YCoor
    right: XCoor
    bottom: YCoor
    left: XCoor

    padding_top: float = 0.0
    padding_right: float = 0.0
    padding_bottom: float = 0.0
    padding_left: float = 0.0

    top_y_value: float = 0.0
    right_x_value: float = 1.0
    bottom_y_value: float = 1.0
    left_x_value: float = 0.0

    def apply_padding(self) -> "Box":
        return Box(
            top=self.top,
            right=self.right,
            bottom=self.bottom,
            left=self.left,
            padding_top=0.0,
            padding_right=0.0,
            padding_bottom=0.0,
            padding_left=0.0,
            top_y_value=self.top_y_value - self.padding_top,
            right_x_value=self.right_x_value + self.padding_right,
            bottom_y_value=self.bottom_y_value + self.padding_bottom,
            left_x_value=self.left_x_value - self.padding_left,
        )


def scale_width_to_box(width: float, box: Box) -> float:
    box = box.apply_padding()
    box_width = box.right_x_value - box.left_x_value
    return abs((width / box_width) * (box.right - box.left))


def scale_height_to_box(height: float, box: Box) -> float:
    box = box.apply_padding()
    box_height = box.bottom_y_value - box.top_y_value
    return abs((height / box_height) * (box.bottom - box.top))


def scale_x_to_box(x: XCoor, box: Box) -> XCoor:
    box = box.apply_padding()
    normalized_x = (x - (box.left_x_value + box.padding_left)) / (
        box.right_x_value - box.left_x_value
    )
    return box.left + normalized_x * (box.right - box.left)


def scale_y_to_box(y: YCoor, box: Box) -> YCoor:
    box = box.apply_padding()
    normalized_y = (y - box.top_y_value) / (box.bottom_y_value - box.top_y_value)
    return box.top + normalized_y * (box.bottom - box.top)


def scale_point_to_box(point: Point, box: Box) -> Point:
    return Point(
        x=scale_x_to_box(point.x, box),
        y=scale_y_to_box(point.y, box),
    )



from pathlib import Path


class SnapshotTest:
    name: str = ""
    path_to_rendered: Path
    description: str = ""

    @staticmethod
    def render() -> str: ...

    @classmethod
    def snapshot_exists(cls) -> bool:
        return cls.path_to_rendered.exists()

    @classmethod
    def load_rendered(cls) -> str:
        if cls.snapshot_exists():
            return cls.path_to_rendered.read_text()
        else:
            return ""

    @classmethod
    def save_rendered(cls, content: str) -> None:
        rendered = cls.render()
        assert rendered == content, "Rendered content does not match provided content."

        cls.path_to_rendered.parent.mkdir(parents=True, exist_ok=True)
        cls.path_to_rendered.write_text(rendered)

    @classmethod
    def passes_test(cls, content: str | None = None) -> bool:
        if not cls.snapshot_exists():
            return False

        snapshot = cls.load_rendered()
        assert snapshot is not None

        rendered = cls.render()
        if content is not None:
            assert rendered == content, (
                "Rendered content does not match provided content."
            )

        return snapshot == rendered


class AllSnapshotsPass(SnapshotTest):
    name: str = ""
    description: str = "All snapshot tests pass!"

    @staticmethod
    def render() -> str:
        return ""

    @classmethod
    def load_rendered(cls) -> str:
        return cls.render()

    @classmethod
    def passes_test(cls, content: str | None = None) -> bool:
        return True



from typing import Callable
from pathlib import Path

from newsflash.app import App, Page
from newsflash.widgets import HTML, Button, Select, ValueDisplay, Notifications
from newsflash.widgets.widgets import Widget

from .all_tests import get_all_tests, get_next_failing_test, get_test_by_name


class OldSnapshotHTML(HTML):
    id: str = "old-snapshot-html"
    html_content: str = get_next_failing_test().load_rendered()


class NewSnapshotHTML(HTML):
    id: str = "new-snapshot-html"
    html_content: str = get_next_failing_test().render()


class TestSelect(Select):
    id: str = "test-select"
    options: list[str] = list(get_all_tests().keys())
    default: Callable[[], str] = lambda: get_next_failing_test().name

    def on_select(
        self,
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        test_description: "TestDescription",
        approve_button: "ApproveButton",
    ) -> list[Widget]:
        assert self.selected is not None
        test_class = get_test_by_name(self.selected)

        old_snapshot.html_content = test_class.load_rendered()
        new_snapshot.html_content = test_class.render()
        test_description.value = test_class.description

        if test_class.passes_test():
            approve_button.label = "Snapshots Already Match"
            approve_button.disabled = True
        else:
            approve_button.label = "Approve Snapshot"
            approve_button.disabled = False

        return [old_snapshot, new_snapshot, test_description, approve_button]


class NextChangedSnapshotButton(Button):
    id: str = "next-changed-snapshot-button"
    label: str = "Next Changed Snapshot"

    def on_click(
        self,
        test_select: TestSelect,
        test_description: "TestDescription",
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        approve_button: "ApproveButton",
    ) -> list[Widget]:
        next_failing_test = get_next_failing_test()

        test_select.selected = next_failing_test.name
        return [
            test_select,
            *test_select.on_select(
                old_snapshot=old_snapshot,
                new_snapshot=new_snapshot,
                test_description=test_description,
                approve_button=approve_button,
            ),
        ]


class TestDescription(ValueDisplay):
    id: str = "test-description"
    label: str = "Test Description"
    value: str = get_next_failing_test().description


class ApproveButton(Button):
    id: str = "approve-button"
    label: str = (
        "Approve Snapshot"
        if not get_next_failing_test().passes_test()
        else "Snapshots Already Match"
    )
    disabled: bool = get_next_failing_test().passes_test()

    def on_click(
        self,
        test_select: TestSelect,
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        notifications: Notifications,
    ) -> list[Widget]:
        assert test_select.selected is not None
        test_class = get_all_tests()[test_select.selected]

        if not test_class.passes_test(new_snapshot.html_content):
            test_class.save_rendered(new_snapshot.html_content)

            # Update the old snapshot content after approval
            old_snapshot.html_content = test_class.load_rendered()
            self.label = "Snapshots Already Match"
            self.disabled = True
            return [self, old_snapshot]

        else:
            notifications.push("Snapshots already match.")
            return [notifications]


page = Page(
    id="",
    path="/",
    title="Snapshot Test Viewer",
    template=("snapshot", "snapshot_test.html"),
    children=[
        TestSelect(),
        NextChangedSnapshotButton(),
        OldSnapshotHTML(),
        NewSnapshotHTML(),
        TestDescription(),
        ApproveButton(),
    ],
)

app = App(
    pages=[page],
    template_folders=[
        ("snapshot", Path(__file__).parent / "templates"),
    ],
)



from typing import Any, Type, Callable, get_type_hints, TypeVar, TYPE_CHECKING
from inspect import signature

from newsflash.svg.element import Element
from newsflash.endpoints.parsers import RequestValues

if TYPE_CHECKING:
    from newsflash.app import Page


class WidgetContainer(Element):
    widget_id: str
    hx_include: list[str] = []
    template: tuple[str, str] = ("widgets", "container.html")

    include_in_context: set[str] = {
        "widget_id",
        "full_path",
        "hx_include",
    }


W = TypeVar("W", bound="Widget")


class Widget(Element):
    hx_include: list[str] = []
    hx_swap_oob: bool = False

    children: list["Widget"] = []
    parent: "Widget | None" = None
    request_values: RequestValues | None = None

    _values_from_request: list[str] = []
    _include_parent: bool = False
    _callback_fn_name: str | None = None
    _callback_fn_on_parent: bool = False

    def _re_init(self: W, update: dict[str, Any] = {}) -> W:
        for k, v in update.items():
            setattr(self, k, v)

        self._post_init()
        return self

    def _post_init(self) -> None:
        self._build_hx_include()
        if self.request_values is not None:
            self._set_values_from_request(self.request_values)

    def get_all_children(
        self,
        type: Type[W],
        request_values: RequestValues | None = None,
    ) -> list[W]:
        children_of_type = [
            widget._re_init(
                update={
                    "request_values": request_values,
                    "parent": self,
                }
            )
            for widget in self.children
            if isinstance(widget, type)
        ]

        if len(self.children) == 0 and isinstance(self, type):
            return []

        for child in self.children:
            child_children = child.get_all_children(
                type=type,
                request_values=request_values,
            )

            children_of_type.extend(child_children)

        return children_of_type

    def get_child_widget(
        self,
        type: Type[W],
        id: str | None = None,
        request_values: RequestValues | None = None,
    ) -> W:
        children_of_type = [
            widget for widget in self.children if isinstance(widget, type)
        ]

        if id is None:
            all_children_of_type = self.get_all_children(
                type=type,
                request_values=request_values,
            )
            if len(all_children_of_type) == 1:
                return all_children_of_type[0]
            else:
                # TODO: distinguish in error message between no or multiple
                # child widgets found.
                raise ValueError(
                    f"Multiple or no widgets of type {type} found. Specify an id."
                )

        id_split = id.split("/")

        if len(id_split) == 1:
            for widget in children_of_type:
                if widget.id == id:
                    widget_copy = widget._re_init(
                        update={
                            "request_values": request_values,
                            "parent": self,
                        }
                    )
                    return widget_copy

        remaining_id = "/".join(id_split[1:])

        for widget in children_of_type:
            if widget.id == id_split[0]:
                widget_copy = widget._re_init(
                    update={
                        "request_values": request_values,
                        "parent": self,
                    }
                )

                widget_copy._post_init()

                return widget_copy.get_child_widget(
                    type=type,
                    id=remaining_id,
                    request_values=request_values,
                )

        raise ValueError(f"Widget not found: {type} with id {id}")

    @property
    def full_path(self) -> str:
        if self.parent is not None:
            full_path = f"{self.parent.full_path}/{self.id}"
        else:
            full_path = self.id

        full_path = full_path.strip("/")

        return full_path

    @property
    def root_widget(self) -> "Page":
        if self.parent is None:
            # assert isinstance(self, Page), "Root widget must be a Page"
            # TODO: how to assert the correct type here?
            return self
        else:
            return self.parent.root_widget

    def get_additional_context(self) -> dict[str, Any]:
        additional_context = super().get_additional_context()

        for child in self.children:
            child.parent = self

        children_instances = [child._re_init() for child in self.children]
        rendered_children = {child.id: child.render() for child in children_instances}

        additional_context.update({"widgets": rendered_children})
        additional_context.update({"widget_list": list(rendered_children.values())})
        additional_context.update({"full_path": self.full_path})

        return additional_context

    def _build_hx_include(self) -> None:
        callback_fn = self._get_callback_fn()
        if callback_fn is None:
            return

        sig = signature(callback_fn)
        parameters = sig.parameters

        type_hints = get_type_hints(callback_fn)

        include_list: list[str] = []
        for param in parameters:
            if param == "args" or param == "kwargs":
                continue

            widget_type = type_hints[param]
            if not issubclass(widget_type, Widget):
                continue

            widget_instance = widget_type()
            include_list.append(f"#{widget_instance.id}")

        if self._include_parent and self.parent is not None:
            include_list.append(f"#{self.parent.id}")

        self.hx_include = include_list

    def _set_value_from_request(self, key: str, inputs: dict[str, Any]) -> None:
        current_value = getattr(self, key, None)
        assert current_value is not None, f"Widget has no attribute '{key}'"

        value_type = type(current_value)
        value = inputs.get(f"{self.id}-{key}", None)

        if value is not None:
            assert isinstance(value, value_type), (
                f"Expected type {value_type} for key '{key}', got {type(value)}"
            )
            setattr(self, key, value)

    def _set_values_from_request(self, inputs: RequestValues) -> None:
        attributes = {
            k: v
            for k, v in inputs.widget_attributes.items()
            if k.startswith(f"{self.id}-")
        }

        for key in self._values_from_request:
            self._set_value_from_request(key, attributes)

        self.request_values = inputs

    def _get_callback_fn(self) -> Callable | None:
        callback_fn_name = self._callback_fn_name
        if callback_fn_name is None:
            return None

        callback_fn = getattr(self, callback_fn_name, None)

        assert callback_fn is not None, (
            f"Widget has no callback function '{callback_fn_name}'"
        )

        return callback_fn

    def _get_callback_inputs(self) -> dict[str, "Widget"]:
        callback_fn = self._get_callback_fn()
        assert callback_fn is not None, "Widget has no callback function"
        assert self.request_values is not None, "Widget has no request values"

        sig = signature(callback_fn)
        parameters = sig.parameters

        type_hints = get_type_hints(callback_fn)

        input_dict = {}
        for param in parameters:
            widget_type = type_hints.get(param, "Unknown")
            if widget_type == "Unknown" or not issubclass(widget_type, Widget):
                continue

            widget = self.root_widget.get_child_widget(
                type=widget_type,
                request_values=self.request_values,
            )

            input_dict[param] = widget

        return input_dict

    def _call_callback(self) -> list["Widget"]:
        callback_fn = self._get_callback_fn()
        assert callback_fn is not None, "Widget has no callback function"

        widgets_to_render = callback_fn(**self._get_callback_inputs())
        return widgets_to_render

    def _render_update(self) -> str:
        self.hx_swap_oob = True
        return self.render()



from .charts import BarChart, LineChart, Histogram
from .inputs import Input, TextArea, Button, Select
from .text import ValueDisplay, Notifications, Paragraph
from .list import List, Grid
from .html import HTML

__all__ = [
    "BarChart",
    "LineChart",
    "Histogram",
    "Input",
    "TextArea",
    "Button",
    "Select",
    "ValueDisplay",
    "Notifications",
    "Paragraph",
    "List",
    "Grid",
    "HTML",
]



import json
from typing import Any
from urllib.parse import urlparse

from starlette.datastructures import FormData, Headers
from pydantic import BaseModel


class RequestValues(BaseModel):
    trigger_element_id: str
    url_path: str
    widget_attributes: dict[str, Any]


def _parse_trigger_element_id(headers: Headers) -> str:
    hx_trigger = headers["hx-trigger"]
    return hx_trigger


def _parse_url_path(headers: Headers) -> str:
    current_url = headers["hx-current-url"]
    parsed_url = urlparse(current_url)
    return parsed_url.path


def _parse_chart_dimensions(body: FormData) -> dict[str, float]:
    dimensions = body.get("dimensions")
    if not isinstance(dimensions, str):
        return {}

    dimensions_dict: dict[str, Any] = json.loads(dimensions)
    dimensions_dict = {k: float(v) for k, v in dimensions_dict.items()}

    return dimensions_dict


def parse_request_values(body: FormData, headers: Headers) -> RequestValues:
    trigger_element_id = _parse_trigger_element_id(headers)
    url_path = _parse_url_path(headers)

    widget_attributes: dict[str, Any] = {}

    for key, value in body.items():
        if key != "dimensions":
            assert isinstance(value, str)
            widget_attributes[key] = value

    chart_dimensions = _parse_chart_dimensions(body)
    widget_attributes.update(chart_dimensions)

    return RequestValues(
        trigger_element_id=trigger_element_id,
        url_path=url_path,
        widget_attributes=widget_attributes,
    )



from newsflash.svg.element import ElementGroup


class SVG(ElementGroup):
    template: tuple[str, str] = ("svg", "svg.svg")
    width: float
    height: float
    hx_swap_oob: bool = False

    include_in_context: set[str] = {
        "id",
        "width",
        "height",
        "hx_swap_oob",
        "classes",
        "styles",
        "attributes",
    }



from typing import Any, TypeVar, Generic, Self, Callable

from newsflash.endpoints.parsers import RequestValues
from .widgets import Widget


W = TypeVar("W", bound=Widget)


class List(Widget, Generic[W]):
    """A widget that displays a list of items."""

    item_type: type[W]
    items: list[W] | None = None
    default: Callable[[], list[W]] | None = None
    template: tuple[str, str] | None = ("widgets", "list.html")

    include_in_context: set[str] = {"id", "hx_include", "hx_swap_oob"}

    def _post_init(self) -> None:
        if self.items is None:
            if self.default is not None:
                self.items = self.default()
            else:
                self.items = []

        super()._post_init()

    def get_list_item_by_id(self, item_id: str) -> W | None:
        if self.items is None:
            return None
        for item in self.items:
            if item.id == item_id:
                return item
        return None

    def get_list_item_by_index(self, index: int) -> W | None:
        if self.items is None:
            return None
        if 0 <= index < len(self.items):
            return self.items[index]
        return None

    @property
    def num_items(self) -> int:
        """Returns the number of items in the list."""
        if self.items is None:
            return 0
        return len(self.items)

    def set_items(self, items: list[W]) -> None:
        self.items = items

    def append_item(self, item: W) -> Self:
        assert self.items is not None, "Items list is not initialized."
        self.items.append(item)
        return self

    def filter_items(self, filter_fn: Callable[[W], bool]) -> Self:
        assert self.items is not None, "Items list is not initialized."
        filtered_items = [item for item in self.items if filter_fn(item)]
        return self.model_copy(update={"items": filtered_items})

    def get_additional_context(self) -> dict[str, Any]:
        assert self.items is not None, "Items list is not initialized."
        for child in self.items:
            child.parent = self

        # TODO: figure out why this assignment complains
        self.children = self.items  # type: ignore

        return super().get_additional_context()

    def _set_values_from_request(self, inputs: RequestValues) -> None:
        super()._set_values_from_request(inputs)
        items: list[W] = []

        item_idx = 0
        while f"{self.id}-{item_idx}-id" in inputs.widget_attributes:
            item_id = inputs.widget_attributes[f"{self.id}-{item_idx}-id"]

            item_widget = self.item_type(
                id=item_id,
                request_values=inputs,
                parent=self,
            )
            item_widget._post_init()

            items.append(item_widget)
            item_idx += 1

        self.items = items

    def get_child_widget(
        self, type: type[W], id: str, request_values: RequestValues | None = None
    ) -> W:
        child_type = self.item_type
        assert issubclass(child_type, type), (
            f"Child type {child_type} is not a subclass of {type}"
        )

        item_id_split = id.split("/")

        widget_instance = child_type(
            id=item_id_split[0],
            request_values=request_values,
            parent=self,
        )
        widget_instance._post_init()

        if len(item_id_split) == 1:
            return widget_instance
        else:
            return widget_instance.get_child_widget(
                type=type,
                id="/".join(item_id_split[1:]),
                request_values=request_values,
            )


class Grid(List[W]):
    """A widget that displays a grid of items."""

    template: tuple[str, str] | None = ("widgets", "grid.html")
    num_columns: int = 2

    include_in_context: set[str] = {"id", "hx_include", "hx_swap_oob", "num_columns"}



from math import ceil, floor
from collections import Counter

from pydantic import BaseModel

from .utils import (
    get_y_label_positions,
    order_of_magnitude,
    smart_round,
    get_step_decimal_places,
)


class AxisConfig(BaseModel):
    label_positions: list[float] | list[int]
    label_values: list[str] | list[int] | list[float]
    decimal_places: int
    min_value: float | int
    max_value: float | int
    multiplier: float = 1

    @property
    def labels_as_str(self) -> list[str]:
        if all(isinstance(label, str) for label in self.label_values):
            return self.label_values  # type: ignore
        else:
            return [f"{num:,.{self.decimal_places}f}" for num in self.label_values]


class AxesConfig(BaseModel):
    x: AxisConfig
    y: AxisConfig


def build_y_axis_config(
    values: list[int] | list[float],
    num_labels: int = 5,
    min_value: float | int | None = None,
) -> AxisConfig:
    label_positions = get_y_label_positions(
        values, divide_by=(num_labels - 1), min_y=min_value
    )

    ooms = [order_of_magnitude(pos) for pos in label_positions]
    multipliers = [pow(10, floor(oom / 3) * 3) for oom in ooms]
    multiplier = Counter(multipliers).most_common(1)[0][0]

    if multiplier != 1:
        label_position_values = [pos / multiplier for pos in label_positions]
    else:
        label_position_values = label_positions

    if all([isinstance(x, float) for x in label_position_values]):
        decimal_places = get_step_decimal_places(label_position_values)  # type: ignore
    else:
        decimal_places = 0

    return AxisConfig(
        label_positions=label_positions,
        label_values=label_position_values,
        decimal_places=decimal_places,
        min_value=min_value if min_value is not None else min(label_positions),
        max_value=max(label_positions),
        multiplier=multiplier,
    )


def build_spaced_x_axis_config(
    values: list[int] | list[float],
    num_labels: int = 8,
    min_value: float | int | None = None,
    max_value: float | int | None = None,
) -> AxisConfig:
    x_step = ceil(len(values) / num_labels)
    label_positions = values[::x_step]

    if all([isinstance(x, float) for x in values]):
        label_positions, decimal_places = smart_round(label_positions)  # type: ignore
    else:
        decimal_places = 0

    return AxisConfig(
        label_positions=label_positions,
        label_values=label_positions,
        decimal_places=decimal_places,
        min_value=min_value if min_value is not None else min(values),
        max_value=max_value if max_value is not None else max(values),
    )


def build_x_axis_config(
    values: list[float] | list[int],
    label_values: dict[float | int, str | int | float],
    min_value: float | int | None = None,
    max_value: float | int | None = None,
) -> AxisConfig:
    return AxisConfig(
        label_positions=list(label_values.keys()),
        label_values=list(label_values.values()),  # type: ignore
        decimal_places=0,
        min_value=min_value if min_value is not None else min(values),
        max_value=max_value if max_value is not None else max(values),
    )


def build_x_axis_config_barchart(labels: list[str]):
    label_positions = list(range(len(labels)))

    return AxisConfig(
        label_positions=label_positions,
        label_values=labels,
        decimal_places=0,
        min_value=0,
        max_value=len(labels) - 1,
    )



from typing import TYPE_CHECKING

from fastapi import Request
from newsflash.templates.templates import template_registry

if TYPE_CHECKING:
    from newsflash.app import Page


def get_page_callback(page: "Page"):
    async def page_endpoint(request: Request) -> str:
        query_params = request.query_params
        query_params_dict: dict[str, list[str]] = {}
        for k in query_params.keys():
            query_params_dict[k] = query_params.getlist(k)

        page_copy = page.model_copy(deep=True)

        page_attrs = page_copy.__class__.__annotations__
        for k, vs in query_params_dict.items():
            # TODO
            # It doesn't make much sense to loop through all values in this case,
            # because only the last value will be used.
            for v in vs:
                if k in page_attrs:
                    attr_type = page_attrs[k]
                    setattr(page_copy, k, attr_type(v))

        page_copy._post_init()
        page_copy.query_params = query_params_dict

        for child in page_copy.children:
            child._post_init()

        rendered_content = page_copy.render()

        page_template = template_registry.get_template("widgets", "index.html")
        return page_template.render(
            request=request,
            title=page_copy.title,
            content=rendered_content,
        )

    return page_endpoint



from pathlib import Path
from fontTools.ttLib import TTFont


def get_text_width(font: TTFont, text: str, font_size: float) -> int:
    units_per_em = font["head"].unitsPerEm  # type: ignore
    cmap = font.getBestCmap()
    hmtx = font["hmtx"]

    glyphs = [cmap[ord(x)] for x in text]
    advance = [hmtx[glyph][0] for glyph in glyphs]  # type: ignore
    total_advance = sum(advance)

    pixels_per_unit = font_size / units_per_em
    return total_advance * pixels_per_unit


def get_text_height(font: TTFont, text: str, font_size: float) -> tuple[float, float]:
    units_per_em = font["head"].unitsPerEm  # type: ignore
    cmap = font.getBestCmap()
    glyf = font["glyf"]

    y_min = 0
    y_max = 0
    glyphs = [cmap[ord(x)] for x in text]

    for glyph in glyphs:
        try:
            if (new_y_min := glyf[glyph].yMin) < y_min:  # type: ignore
                y_min = new_y_min
            if (new_y_max := glyf[glyph].yMax) > y_max:  # type: ignore
                y_max = new_y_max
        except AttributeError:
            continue

    pixels_per_unit = font_size / units_per_em
    return y_min * pixels_per_unit, y_max * pixels_per_unit


def font_size_to_height(font_size: int) -> float:
    return font_size * 4 / 3


lora = TTFont(
    str(
        Path(__file__).parent.parent.parent
        / "assets"
        / "fonts"
        / "lora"
        / "Lora-Regular.ttf"
    )
)



from fontTools.ttLib import TTFont

from newsflash.svg.box import Box
from newsflash.svg.utils import Point
from newsflash.svg.utils.fonts import lora
from newsflash.svg.element import ElementGroup
from newsflash.svg.elements import build_path

from .xy_chart import build_xy_chart
from .axes import (
    AxesConfig,
    build_x_axis_config,
    build_spaced_x_axis_config,
    build_y_axis_config,
)


def _build_line(
    xs: list[float] | list[int],
    ys: list[float] | list[int],
    chart_box: Box,
) -> ElementGroup:
    elements = ElementGroup()

    line_points = [
        Point(
            x=x,
            y=y,
        )
        for x, y in list(zip(xs, ys))
    ]

    path = build_path(
        points=line_points,
        box=chart_box,
        classes=["line-path"],
    )

    elements.append(path)
    return elements


def build_linechart(
    xs: list[float] | list[int],
    ys: list[float] | list[int],
    width: float,
    height: float,
    title: str,
    x_labels: dict[float | int, str | int | float] | None = None,
    min_x_value: float | int | None = None,
    max_x_value: float | int | None = None,
    font: TTFont = lora,
    title_font_size: int = 32,
    label_font_size: int = 16,
) -> ElementGroup:
    linechart_elements = ElementGroup()

    x_padding = (max(xs) - min(xs)) / 50

    if x_labels is not None:
        x_axis_config = build_x_axis_config(
            values=xs,
            label_values=x_labels,
            min_value=min_x_value,
            max_value=max_x_value,
        )
    else:
        x_axis_config = build_spaced_x_axis_config(
            values=xs,
            min_value=min_x_value,
            max_value=max_x_value,
        )

    axes = AxesConfig(
        x=x_axis_config,
        y=build_y_axis_config(values=ys),
    )

    linechart_elements, chart_box = build_xy_chart(
        axes=axes,
        width=width,
        height=height,
        title=title,
        x_padding=x_padding,
        font=font,
        title_font_size=title_font_size,
        label_font_size=label_font_size,
    )

    line = _build_line(xs=xs, ys=ys, chart_box=chart_box)
    linechart_elements.extend(line)

    return linechart_elements
